import datetime as dt
import itertools
import logging
import re
from collections.abc import Sequence
from pathlib import Path
from textwrap import dedent, indent
from typing import (
    Annotated,
    Any,
    Literal,
    Optional,
    Union,
    get_args,
    get_origin,
    no_type_check,
)

import cftime  # noqa: F401
import xarray as xr
from pydantic import (
    ConfigDict,
    Field,
    FilePath,
    HttpUrl,
    NonNegativeFloat,
    PositiveInt,
    PrivateAttr,
    ValidationError,
    field_validator,
    model_validator,
)

from ..config import options
from .base import (
    Command,
    FlatCommand,
    GenericParameterList,
    LineCommand,
    ListCommand,
    ParameterList,
    Record,
    RootCommand,
    RootRecord,
    Sym,
)
from .utils import filter_for, get_annotations, nc_specs


LOGGER = logging.getLogger("RavenPy")


"""
Raven commands
==============

The syntax of those commands match as closely as possible the Raven documentation.
"""


# TODO: Strip command outputs of white space to facilitate testing.
# TODO: Add docstrings
# TODO: Create tests in tests/config/test_commands.py


INDENT = " " * 4
VALUE_PADDING = 10
T12 = tuple[
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
]


class Process(Command):
    """
    Process type embedded in HydrologicProcesses command.

    See processes.py for list of processes.
    """

    algo: str = "RAVEN_DEFAULT"
    source: Optional[str] = None
    to: Sequence[str] = ()

    _sub = ""

    @field_validator("to", mode="before")
    @classmethod
    def is_list(cls, v):
        if isinstance(v, str):
            return [
                v,
            ]
        return v

    @field_validator("source")
    @classmethod
    def is_source_state_variable(cls, v: str):
        if v is not None:
            if re.split(r"\[", v)[0] not in get_args(options.StateVariables):
                raise ValueError(f"{v} not recognized as a state variable.")
        return v

    @field_validator("to")
    @classmethod
    def is_to_state_variable(cls, v: Sequence):
        for x in v:
            if re.split(r"\[", x)[0] not in get_args(options.StateVariables):
                raise ValueError(f"{x} not recognized as a state variable.")
        return v

    def to_rv(self):
        t = " ".join(map("{:<19}".format, self.to))
        p = getattr(self, "p", None)
        p = p if p is not None else ""
        cmd = self._sub + self.__class__.__name__
        return f":{cmd:<20} {self.algo:<19} {self.source:<19} {t} {p}".strip()


class SoilModel(RootCommand):
    """:SoilModel SOIL_MULTILAYER 6"""

    root: int

    def to_rv(self):
        cmd = "SoilModel"
        return f":{cmd:<20} SOIL_MULTILAYER {self.root}\n"


class LinearTransform(Command):
    """:LinearTransform 1.0 -273.15"""

    scale: float = 1
    offset: float = 0

    def to_rv(self):
        cmd = "LinearTransform"
        if (self.scale != 1) or (self.offset != 0):
            return f":{cmd:<20} {self.scale:.15f} {self.offset:.15f}\n"
        return ""


class RainSnowTransition(Command):
    """
    Specify the range of temperatures over which there will be a rain/snow mix when partitioning total precipitation into rain/snow components.

    :RainSnowTransition [temp] [delta]

    # equivalent to (the preferred option)
    :GlobalParameter RAINSNOW_TEMP [rainsnow_temp]
    :GlobalParameter RAINSNOW_DELTA [rainsnow_delta]
    """

    temp: Sym
    """Midpoint of the temperature range [C]."""
    delta: Sym
    """Range [C]."""

    def to_rv(self):
        cmd = "RainSnowTransition"
        return f":{cmd:<20} {self.temp} {self.delta}\n"


class EvaluationPeriod(LineCommand):
    """:EvaluationPeriod [period_name] [start yyyy-mm-dd] [end yyyy-mm-dd]"""

    name: str
    start: dt.date
    end: dt.date


class CustomOutput(LineCommand):
    """
    Create custom output file to track a single variable, parameter, or forcing function over time at a number of basins, HRUs, or across watershed.

    :CustomOutput DAILY AVERAGE AET BY_HRU
    """  # noqa: E501

    time_per: Literal["DAILY", "MONTHLY", "YEARLY", "WATER_YEARLY", "CONTINUOUS"]
    """Time period."""
    stat: Literal["AVERAGE", "MAXIMUM", "MINIMUM", "RANGE", "MEDIAN", "QUARTILES"]
    """Statistic reported for each time interval."""
    variable: str
    """
    Variable or parameter name.

    Consult the Raven documentation for the list of allowed names.
    """
    space_agg: Literal["BY_BASIN", "BY_HRU", "BY_HRU_GROUP", "BY_SB_GROUP", "ENTIRE_WATERSHED"]
    """Spatial evaluation domain."""
    filename: str = ""
    """
    Output file name.

    Defaults to something approximately like: `<run name>_<variable>_<time_per>_<stat>_<space_agg>.nc`.
    """


class SoilProfile(Record):
    name: str = ""
    soil_classes: Sequence[str] = ()
    thicknesses: Sequence[Sym] = ()

    def __str__(self):
        # From the Raven manual: {profile_name,#horizons,{soil_class_name,thick.}x{#horizons}}x[NP]
        n_horizons = len(self.soil_classes)
        horizon_data = list(itertools.chain(*zip(self.soil_classes, self.thicknesses)))
        fmt = "{:<16},{:>4}," + ",".join(n_horizons * ["{:>12}, {:>8}"])
        return fmt.format(self.name, n_horizons, *horizon_data)


class SoilProfiles(ListCommand):
    """
    SoilProfiles command.

    Example::

        :SoilProfiles
          # name, #horizons, hor1, th1, hor2, th2
          LAKE, 0
          GLACIER, 0
          LOAM_SEQ, 2, LOAM, 0.5, SAND, 1.5
          ALL_SAND, 2, SAND, 0.5, SAND, 1.5
        :EndSoilProfiles
    """

    root: Sequence[SoilProfile]


class SubBasin(Record):
    """Record to populate RVH :SubBasins command internal table."""

    subbasin_id: PositiveInt = 1
    name: str = "sub_001"
    downstream_id: int = -1
    profile: str = "NONE"
    reach_length: float | str = 0
    gauged: bool = True
    gauge_id: Optional[str] = ""  # This attribute is not rendered to RVH

    @classmethod
    @field_validator("reach_length")
    def check_reach_length(cls, v):
        if v == "ZERO-":
            return 0
        return v

    def __str__(self):
        d = self.model_dump()
        d["gauged"] = int(d["gauged"])
        del d["gauge_id"]
        return " ".join(f"{v: <{VALUE_PADDING}}" for v in d.values())

    model_config = ConfigDict(coerce_numbers_to_str=True)


class SubBasins(ListCommand):
    """
    SubBasins command (RVH)

    Example::

        :SubBasins
          :Attributes, NAME, DOWNSTREAM_ID, PROFILE, REACH_LENGTH, GAUGED
          :Units, none, none, none, km, none
          1, Downstream, -1, DEFAULT, 3.0, 1
          2, Upstream, 1, DEFAULT, 3.0, 0
        :EndSubBasins
    """

    root: Sequence[SubBasin]
    _Attributes: Sequence[str] = PrivateAttr(["ID", "NAME", "DOWNSTREAM_ID", "PROFILE", "REACH_LENGTH", "GAUGED"])
    _Units: Sequence[str] = PrivateAttr(["none", "none", "none", "none", "km", "none"])

    @classmethod
    def parse(cls, s):
        """Parse a SubBasins command."""
        pat = r":SubBasins(.+):EndSubBasins" ""
        out = []
        keys = [
            "subbasin_id",
            "name",
            "downstream_id",
            "profile",
            "reach_length",
            "gauged",
        ]

        # Remove all commented lines (starting with #)
        s = re.sub(r"^\s*#.*$", "", s, flags=re.MULTILINE)
        if match := re.search(pat, s.strip(), re.DOTALL):
            spat = r"^\s*([^:^\s].+)$"
            for line in re.findall(spat, match.groups()[0], re.MULTILINE):
                # Split line into fields
                values = re.split(r"\s+", line.strip())
                # Convert to SubBasin record
                out.append(SubBasin(**dict(zip(keys, values))))
        return cls(root=out)


class HRU(Record):
    """Record to populate :HRUs command internal table (RVH)."""

    hru_id: PositiveInt = 1
    area: Sym = 0  # km^2
    elevation: float = 0  # meters
    latitude: float = 0
    longitude: float = 0
    subbasin_id: PositiveInt = 1
    land_use_class: str = "[NONE]"
    veg_class: str = "[NONE]"
    soil_profile: str = "[NONE]"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    slope: NonNegativeFloat = 0.0
    aspect: Annotated[float, Field(ge=0, le=360)] = 0.0
    # This field is not part of the Raven config, it is needed for serialization,
    # to specify which HRU subclass to use when necessary
    hru_type: Optional[str] = None

    def __str__(self):
        import numpy as np

        d = self.model_dump()
        del d["hru_type"]

        # Adjust horizontal spacing to match attributes names
        attrs = ["HRU_ID"] + HRUs._Attributes.default
        ns = np.array(list(map(len, attrs)))
        ns[1:] += 1

        # In Py3.10 zip(strict=True) can be used instead.
        if len(d) != len(attrs):
            raise ValueError(f"HRU record has {len(d)} attributes, but {len(attrs)} are expected.")
        return " ".join(f"{v:<{n}}" for v, n in zip(d.values(), ns))


class HRUs(ListCommand):
    """HRUs command (RVH)."""

    root: Sequence[HRU]
    _Attributes: Sequence[str] = PrivateAttr(
        [
            "AREA",
            "ELEVATION",
            "LATITUDE",
            "LONGITUDE",
            "BASIN_ID",
            "LAND_USE_CLASS",
            "VEG_CLASS",
            "SOIL_PROFILE",
            "AQUIFER_PROFILE",
            "TERRAIN_CLASS",
            "SLOPE",
            "ASPECT",
        ]
    )
    _Units: Sequence[str] = PrivateAttr(
        [
            "km2",
            "m",
            "deg",
            "deg",
            "none",
            "none",
            "none",
            "none",
            "none",
            "none",
            "deg",
            "degN",
        ]
    )

    @field_validator("root", mode="before")
    @classmethod
    def ignore_unrecognized_hrus(cls, values):
        """
        Ignore HRUs with unrecognized hru_type.

        HRUs are ignored only if all allowed HRU classes define `hru_type`, and if the values passed include it.
        """
        import collections
        import warnings

        a = cls.model_fields["root"].annotation

        # Annotation should be a sequence
        if get_origin(a) != collections.abc.Sequence:
            return values

        # Extract allowed HRU types
        allowed = [hru.model_fields["hru_type"].default for hru in get_annotations(a)]

        # If some HRU classes do not define rhu_type, skip filtering
        if None in allowed:
            return values

        allowed.append(None)

        out = [value for value in values if getattr(value, "hru_type", None) in allowed]
        if len(out) != len(values):
            warnings.warn("HRUs with an unrecognized `hru_type` attribute were ignored.", stacklevel=2)
        return out

    @classmethod
    def parse(cls, s):
        """Parse a HRUs command."""
        pat = r":HRUs(.+):EndHRUs"
        out = []
        keys = [
            "hru_id",
            "area",
            "elevation",
            "latitude",
            "longitude",
            "subbasin_id",
            "land_use_class",
            "veg_class",
            "soil_profile",
            "aquifer_profile",
            "terrain_class",
            "slope",
            "aspect",
        ]
        # Remove all commented lines (starting with #)
        s = re.sub(r"^\s*#.*$", "", s, flags=re.MULTILINE)

        if match := re.search(pat, s.strip(), re.DOTALL):
            spat = r"^\s*([^:^\s].+)$"
            for line in re.findall(spat, match.groups()[0], re.MULTILINE):
                # Split line into fields
                values = re.split(r"\s+", line.strip())
                # Convert to HRU record
                out.append(dict(zip(keys, values)))
        return cls(root=out)

    def rename(self, mapping):
        """
        Rename HRU attributes.

        Parameters
        ----------
        mapping : dict
            Nested dictionary keyed by HRU attributes, with keys as old names and values as new names.
        """
        out = self.copy()
        for attr, amap in mapping.items():
            for hru in out.root:
                cur = getattr(hru, attr)
                for old, new in amap.items():
                    if cur == old:
                        setattr(hru, attr, new)
        return out


class HRUGroup(FlatCommand):
    class _Rec(RootRecord):
        root: Sequence[str]

        def __str__(self):
            return ",".join(self.root)

    name: str
    groups: _Rec

    @property
    def _template(self):
        return """
        :{_cmd} {name}
        {_records}
        :End{_cmd}
        """


class Reservoir(FlatCommand):
    """Reservoir command (RVH)."""

    name: str = "Lake_XXX"
    subbasin_id: int = Field(0, alias="SubBasinID")
    hru_id: int = Field(0, alias="HRUID")
    type: str = Field("RESROUTE_STANDARD", alias="Type")
    weir_coefficient: float = Field(0, alias="WeirCoefficient")
    crest_width: float = Field(0, alias="CrestWidth")
    max_depth: float = Field(0, alias="MaxDepth", description="Max depth (m)")
    lake_area: float = Field(0, alias="LakeArea", description="Lake area (m2)")
    absolute_crest_height: Optional[float] = Field(None, alias="AbsoluteCrestHeight", description="Absolute crest height (m)")
    max_capacity: Optional[float] = Field(None, alias="MaxCapacity", description="Maximum capacity in m3")

    class SeepageParameters(LineCommand):
        """:SeepageParameters [K_seep] [href]"""

        k_seep: float
        h_ref: float

    seepage_parameters: Optional[SeepageParameters] = Field(None, alias="SeepageParameters")

    class StageRelations(RootCommand):
        """Stage relations for the reservoir."""

        class StageRelation(RootRecord):
            """
            Stage relation record.

            h, q, v, a, [u]
            """

            root: tuple[float, float, float, float] | tuple[float, float, float, float, float] = None

            def __str__(self):
                return ", ".join([f"{x:>12}" for x in self.root])

        root: Sequence[StageRelation] = Field(None)

    stage_relations: Optional[StageRelations] = Field(None, alias="StageRelations")

    class OutflowControlStructure(Command):
        """Outflow control structure for the reservoir."""

        target_subbasin_id: Optional[int] = Field(None, alias="TargetSubBasin")
        downstream_reference_elevation: Optional[float] = Field(None, alias="DownstreamReferenceElevation")

        class StageDischargeTable(RootCommand):
            """
            Stage discharge table for the outflow control structure.

            Example::

                :StageDischargeTable C1 #one gate open
                      N
                      [h,Q]xN
                :EndStageDischargeTable
            """

            _n: Optional[int] = None

            class StageDischargeRecord(RootRecord):
                """Stage discharge record."""

                root: tuple[float, float] = None

            root: Sequence[StageDischargeRecord] = Field(None)

            @model_validator(mode="after")
            def _set_n(self):
                """Set the number of records in the table."""
                if self.root is not None and self.n is None:
                    self.n = len(self.root)
                return self

            @property
            def _template(self):
                return """
                :StageDischargeTable
                  {n}
                {_records}
                :EndStageDischargeTable
                """

        stage_discharge_table: Optional[Sequence[StageDischargeTable]] = Field(None, alias="StageDischargeTable")

        class BasicWeir(LineCommand):
            """
            Basic weir for the outflow control structure.

            :BasicWeir [curve_name3] [elev] [crestwidth] [coeff]
            """

            curve_name: str
            elev: float
            crest_width: float
            coeff: float

        basic_weir: BasicWeir = Field(None, alias="BasicWeir")

        class OperatingRegime(Command):
            """
            Operating regime for the outflow control structure.

            :OperatingRegime [curve_name] [elev] [coeff]
            """

            name: str
            use_curve: str = Field(None, alias="UseCurve")
            condition: str = Field(None, alias="Condition")
            constraint: str = Field(None, alias="Constraint")

        operating_regime: OperatingRegime = Field(None, alias="OperatingRegime")

    @property
    def _template(self):
        return """
            :Reservoir {name}
            {_commands}
            :EndReservoir
            """

    @classmethod
    def parse(cls, s) -> list:
        """Return a list of Reservoir records parsed from an RV file."""
        # Remove all commented lines (starting with #)
        s = re.sub(r"^\s*#.*$", "", s, flags=re.MULTILINE)
        pat = r":Reservoir\s+(\w+)\s+(.+?):EndReservoir"
        out = []

        for name, content in re.findall(pat, s.strip(), re.DOTALL):
            # Extract parameters
            subbasin_id = re.search(r":SubBasinID\s+(\d+)", content).group(1)
            hru_id = re.search(r":HRUID\s+(\d+)", content).group(1)
            type_ = re.search(r":Type\s+(\w+)", content).group(1)
            weir_coefficient = re.search(r":WeirCoefficient\s+([\d.-]+)", content).group(1)
            crest_width = re.search(r":CrestWidth\s+([\d.-]+)", content).group(1)
            max_depth = re.search(r":MaxDepth\s+([\d.-]+)", content).group(1)
            lake_area = re.search(r":LakeArea\s+([\d.-]+)", content).group(1)
            seepage_parameters = re.search(r":SeepageParameters\s+([\d.-]+)\s+([\d.-]+)", content).groups()

            # Convert to Reservoir record
            out.append(
                cls(
                    name=name,
                    subbasin_id=subbasin_id,
                    hru_id=hru_id,
                    type=type_,
                    weir_coefficient=weir_coefficient,
                    crest_width=crest_width,
                    max_depth=max_depth,
                    lake_area=lake_area,
                    seepage_parameters=cls.SeepageParameters(
                        k_seep=seepage_parameters[0],
                        h_ref=seepage_parameters[1],
                    ),
                )
            )
        return out


class SubBasinGroup(FlatCommand):
    """SubBasinGroup command (RVH)."""

    name: str = ""
    sb_ids: Sequence[int] = ()

    def to_rv(self):
        template = """
            :SubBasinGroup {name}
              {sb_ids}
            :EndSubBasinGroup
            """
        d = self.model_dump()
        n_per_line = 10
        sbids = sorted(self.sb_ids)
        sbids_lines = [map(str, sbids[i : i + n_per_line]) for i in range(0, len(sbids), n_per_line)]
        d["sb_ids"] = "\n  ".join([", ".join(sbids) for sbids in sbids_lines])
        return dedent(template).format(**d)

    @classmethod
    def parse(cls, s) -> list:
        """Parse a SubBasinGroup command and return a list of SubBasinGroup records."""
        # Remove all commented lines (starting with #)
        s = re.sub(r"^\s*#.*$", "", s, flags=re.MULTILINE)
        pat = r":SubBasinGroup\s+(\w+)\s+(.+?):EndSubBasinGroup"
        out = []

        for name, content in re.findall(pat, s.strip(), re.DOTALL):
            sb_ids = re.split(r"\s+", content.strip())

            # Convert to SubBasinGroup record
            out.append(SubBasinGroup(name=name, sb_ids=sb_ids))
        return out


class SBGroupPropertyMultiplier(LineCommand):
    """:SBGroupPropertyMultiplier [group_name] [parameter_name] [mult]"""

    group_name: str
    parameter_name: str
    mult: Sym


class ChannelProfile(FlatCommand):
    """ChannelProfile command (RVP)."""

    name: str = "chn_XXX"
    bed_slope: float = Field(0, alias="Bedslope")

    class SurveyPoints(ListCommand):
        """
        SurveyPoints

        [x, bed_elevation] x number of survey points.
        """

        class SurveyPoint(RootRecord):
            """SurveyPoint record."""

            root: tuple[float, float] = ()

        root: Sequence[SurveyPoint] = Field((SurveyPoint(),))

    survey_points: SurveyPoints = Field(SurveyPoints(), alias="SurveyPoints")

    class RoughnessZones(ListCommand):
        """
        RoughnessZones record.

        [x_zone, mannings_n] x number of roughness zones.
        """

        class RoughnessZone(RootRecord):
            """RoughnessZone record."""

            root: tuple[float, float] = ()

        root: Sequence[RoughnessZone] = Field((RoughnessZone(),))

    roughness_zones: RoughnessZones = Field(RoughnessZones(), alias="RoughnessZones")

    @property
    def _template(self):
        return """
               :{_cmd} {name}
               {_commands}{_records}
               :End{_cmd}
               """

    @classmethod
    def parse(cls, s) -> list:
        """Parse ChannelProfile commands and return a list of ChannelProfile records."""
        # Remove all commented lines (starting with #)
        s = re.sub(r"^\s*#.*$", "", s, flags=re.MULTILINE)

        pat = r":ChannelProfile\s+(\w+)\s+(.+?):EndChannelProfile"
        out = []

        for name, content in re.findall(pat, s.strip(), re.DOTALL):
            bed_slope = re.search(r":Bedslope\s+([\d.-]+)", content).group(1)
            survey_points = re.search(r":SurveyPoints(.+):EndSurveyPoints", content, re.DOTALL).group(1)
            sp = [line.strip().split() for line in survey_points.splitlines() if line.strip()]

            roughness_zones = re.search(r":RoughnessZones(.+):EndRoughnessZones", content, re.DOTALL).group(1)
            rz = [line.split() for line in roughness_zones.splitlines() if line.strip()]
            # rz = re.split(r"\s+", roughness_zones.strip())

            # Convert to ChannelProfile record
            out.append(
                cls(
                    name=name,
                    bed_slope=bed_slope,
                    survey_points=sp,
                    roughness_zones=rz,
                )
            )
        return out


class GridWeights(Command):
    """
    GridWeights command.

    Notes
    -----
    command can be embedded in both a `GriddedForcing` or a `StationForcing`.

    The default is to have a single cell that covers an entire single HRU, with a
    weight of 1.
    """

    number_hrus: int = Field(1, alias="NumberHRUs")
    number_grid_cells: int = Field(1, alias="NumberGridCells")

    class GWRecord(RootRecord):
        """GridWeights record."""

        root: tuple[int, int, float] = (1, 0, 1.0)

    data: Sequence[GWRecord] = Field((GWRecord(),))

    # @field_validator("data")
    # @classmethod
    # def _sum_weights(cls, values):
    #     """Check that for each HRU weights sum to 1."""
    #     for hru, group in itertools.groupby(values, lambda x: x.root[0]):
    #         total = sum([x.root[2] for x in group])
    #         if not (0.999 < total < 1.001):
    #             raise ValueError(
    #                 f"GridWeights for HRU {hru} do not sum to 1.0: {total:.3f}"
    #             )
    #     return values

    @classmethod
    def parse(cls, s):
        pat = r"""
        :GridWeights
        \s*:NumberHRUs\s+(\d+)
        \s*:NumberGridCells\s+(\d+)
        \s*(.+)
        :EndGridWeights
        """
        m = re.match(dedent(pat).strip(), s.strip(), re.DOTALL)
        n_hrus, n_grid_cells, data = m.groups()
        data = [d.strip().split() for d in data.split("\n")]
        data = tuple((int(h), int(c), float(w)) for h, c, w in data)
        return cls(number_hrus=int(n_hrus), number_grid_cells=int(n_grid_cells), data=data)


class RedirectToFile(RootCommand):
    """
    RedirectToFile command (RVT).

    Notes
    -----
    For the moment, this command can only be used in the context of a `GriddedForcingCommand`
    or a `StationForcingCommand`, as a `grid_weights` field replacement when inlining is not
    desired.
    """

    root: FilePath

    def to_rv(self):
        cmd = "RedirectToFile"
        return f":{cmd:<20} {self.root}\n"


class ReadFromNetCDF(FlatCommand):
    # TODO: When encoding Path, return only path.name
    # Order of HttpUrl, Path is important to avoid casting strings to Path.
    file_name_nc: Union[HttpUrl, Path] = Field(..., alias="FileNameNC", description="NetCDF file name.")
    var_name_nc: str = Field(..., alias="VarNameNC", description="NetCDF variable name.")
    dim_names_nc: Sequence[str] = Field(..., alias="DimNamesNC")
    station_idx: int = Field(
        1,
        alias="StationIdx",
        description="NetCDF index along station dimension. Starts at 1.",
    )
    time_shift: Optional[float] = Field(None, alias="TimeShift", description="Time stamp shift in days.")
    linear_transform: Optional[LinearTransform] = Field(None, alias="LinearTransform")
    deaccumulate: Optional[bool] = Field(None, alias="Deaccumulate")

    latitude_var_name_nc: Optional[str] = Field(None, alias="LatitudeVarNameNC")
    longitude_var_name_nc: Optional[str] = Field(None, alias="LongitudeVarNameNC")
    elevation_var_name_nc: Optional[str] = Field(None, alias="ElevationVarNameNC")

    @field_validator("dim_names_nc")
    @classmethod
    def reorder_time(cls, v):
        """TODO: Return dimensions as x, y, t. Currently only puts time at the end."""
        dims = list(v)
        for time_dim in ("t", "time"):
            if time_dim in dims:
                dims.remove(time_dim)
                dims.append(time_dim)
                break
        else:
            raise ValidationError("No time dimension found in dim_names_nc tuple.")

        return tuple(dims)

    @classmethod
    def from_nc(cls, fn, data_type, station_idx=None, alt_names=(), engine="h5netcdf", **kwds):
        """Instantiate class from netCDF dataset."""
        specs = nc_specs(
            fn,
            data_type,
            station_idx=station_idx,
            alt_names=alt_names,
            engine=engine,
        )
        specs.update(kwds)
        attrs = filter_for(cls, specs)
        return cls(**attrs)

    @property
    def da(self) -> xr.DataArray:
        """Return DataArray from configuration."""
        # TODO: Apply linear transform and time shift
        da = xr.open_dataset(self.file_name_nc)[self.var_name_nc]
        if len(self.dim_names_nc) == 1:
            return da
        elif len(self.dim_names_nc) == 2:
            return da.isel({self.dim_names_nc[0]: self.station_idx})
        else:
            raise NotImplementedError()


class GriddedForcing(ReadFromNetCDF):
    """GriddedForcing command (RVT)."""

    name: str = ""
    forcing_type: Optional[options.Forcings] = Field(None, alias="ForcingType")
    grid_weights: Union[GridWeights, RedirectToFile] = Field(GridWeights(), alias="GridWeights")
    # StationIdx is not relevant to GriddedForcing
    station_idx: Optional[int] = None

    @property
    def _template(self):
        return """
        :{_cmd} {name}
        # HRU GridCell Weight
        {_commands}
        :End{_cmd}
        """

    @field_validator("dim_names_nc")
    @classmethod
    def check_dims(cls, v):
        if len(v) != 3:
            raise ValueError("GriddedForcing netCDF datasets should have  three dimensions (lon, lat, time).")
        return v


class StationForcing(GriddedForcing):
    """StationForcing command (RVT)."""

    @field_validator("dim_names_nc")
    @classmethod
    def check_dims(cls, v):
        if len(v) != 2:
            raise ValueError("StationForcing netCDF datasets should have two dimensions (station, time).")
        return v

    @classmethod
    def from_nc(cls, fn, data_type, alt_names=(), **kwds):
        """Instantiate class from netCDF dataset."""
        specs = nc_specs(fn, data_type, station_idx=None, alt_names=alt_names)
        specs.update(kwds)

        attrs = filter_for(cls, specs)

        # Build default gridweights
        if "grid_weights" not in kwds and "GridWeights" not in kwds:
            size = specs["_dim_size_nc"]
            size.pop(specs["_time_dim_name_nc"])
            nstations = list(size.values())[0]
            data = [[i + 1, i + 1, 1] for i in range(nstations)]
            gw = GridWeights(number_hrus=nstations, number_grid_cells=nstations, data=data)
            attrs["grid_weights"] = gw

        return cls(**attrs)


class Data(FlatCommand):
    data_type: options.Forcings = ""
    units: str = ""
    read_from_netcdf: ReadFromNetCDF = Field(..., alias="ReadFromNetCDF")

    @property
    def _template(self):
        return """
               :{_cmd} {data_type} {units}
               {_commands}
               :End{_cmd}
               """

    _nested: bool = False

    @classmethod
    def from_nc(cls, fn, data_type, station_idx=1, alt_names=(), **kwds):
        specs = nc_specs(fn, data_type, station_idx, alt_names)
        specs.update(kwds)
        attrs = filter_for(cls, specs)
        return cls(**attrs)


class Gauge(FlatCommand):
    name: str = "default"
    latitude: float = Field(..., alias="Latitude")
    longitude: float = Field(..., alias="Longitude")
    elevation: Optional[float] = Field(None, alias="Elevation")

    rain_correction: Optional[Sym] = Field(None, alias="RainCorrection", description="Rain correction")
    snow_correction: Optional[Sym] = Field(None, alias="SnowCorrection", description="Snow correction")

    monthly_ave_evaporation: Optional[Sequence] = Field(None, alias="MonthlyAveEvaporation")
    monthly_ave_temperature: Optional[Sequence] = Field(None, alias="MonthlyAveTemperature")
    monthly_min_temperature: Optional[Sequence] = Field(None, alias="MonthlyMinTemperature")
    monthly_max_temperature: Optional[Sequence] = Field(None, alias="MonthlyMaxTemperature")

    data: Optional[Sequence[Data]] = Field(None, alias="Data")

    _nested: bool = False

    @property
    def _template(self):
        return """
        :Gauge {name}
        {_commands}
        :EndGauge
        """

    @field_validator("monthly_ave_evaporation", "monthly_ave_temperature")
    @classmethod
    def confirm_monthly(cls, v):
        if v is not None and len(v) != 12:
            raise ValidationError("One value per month needed.")
        return v

    @classmethod
    def from_nc(
        cls,
        fn: Union[str, Path, Sequence[Path]],
        data_type: Optional[Sequence[str]] = None,
        station_idx: int = 1,
        alt_names: Optional[dict[str, str]] = None,
        mon_ave: bool = False,
        data_kwds: Optional[dict[str, Any]] = None,
        engine: str = "h5netcdf",
        **kwds,
    ) -> "Gauge":
        r"""
        Return Gauge instance with configuration options inferred from the netCDF itself.

        Parameters
        ----------
        fn : str or Path or Sequence[Path]
            NetCDF file path or paths.
        data_type : Sequence[str], optional
            Raven data types to extract from netCDF files, e.g. 'PRECIP', 'AVE_TEMP'. The algorithm tries to find all
            forcings in each file until one is found, then it stops searching for it in the following files.
        station_idx : int
            Index along station dimension. Starts at 1. Should be the same for all netCDF files.
        alt_names : dict
            Alternative variable names keyed by data type.
            Use this if variables do not correspond to CF standard defaults.
        mon_ave : bool
            If True, compute the monthly average.
        data_kwds : dict[options.Forcings, dict[str, str]]
            Additional `:Data` parameters keyed by forcing type and station id. Overrides inferred parameters.
            Use keyword "ALL" to pass parameters to all variables.
        engine : {"h5netcdf", "netcdf4", "pydap"}
            The engine used to open the dataset. Default is 'h5netcdf'.
        \*\*kwds : dict
            Additional arguments for Gauge.

        Returns
        -------
        Gauge
            Gauge instance.
        """
        if alt_names is None:
            alt_names = {}
        if data_kwds is None:
            data_kwds = {}

        if isinstance(fn, (str, Path)):
            fn = [fn]

        idx = station_idx
        data = {}
        attrs = {}

        forcings = set(data_type or get_args(options.Forcings))

        for f in fn:
            for dtype in forcings:
                try:
                    specs = nc_specs(
                        fn=f,
                        data_type=dtype,
                        station_idx=idx,
                        alt_names=alt_names.get(dtype, ()),
                        mon_ave=mon_ave,
                        engine=engine,
                    )
                except ValueError as e:  # noqa: PERF203
                    LOGGER.info(e)
                    pass

                else:
                    # Add extra keywords
                    ex_f = {**data_kwds.get(dtype, {}), **data_kwds.get("ALL", {})}

                    # Filter data attributes
                    data[dtype] = filter_for(Data, specs, **ex_f)

                    attrs.update(filter_for(cls, specs))

            # Remove forcings that have been found, so that they're not searched in the following files.
            forcings.difference_update(data)

        if len(data) == 0:
            raise ValueError(
                "No data found in netCDF files. Check that variable names follow CF conventions, "
                "or if not, provide `alt_names` mapping Raven data types to variable names."
            )

        # Default Gauge name
        attrs["name"] = attrs.get("name", f"Gauge_{idx}")

        attrs = filter_for(cls, attrs, **kwds, data=list(data.values()))
        return cls(**attrs)

    @property
    def ds(self) -> xr.Dataset:
        """Return xarray Dataset with forcing variables keyed by Raven forcing names."""
        ds = {}
        for data in self.data:
            ds[data.data_type] = data.read_from_netcdf.da
        return xr.Dataset(ds)


class ObservationData(Data, coerce_numbers_to_str=True):
    uid: str = "1"  # Basin ID or HRU ID
    data_type: Literal["HYDROGRAPH"] = "HYDROGRAPH"

    @property
    def _template(self):
        return """
        :ObservationData {data_type} {uid} {units}
        {_commands}
        :EndObservationData
        """

    @classmethod
    def from_nc(cls, fn, station_idx: int = 1, alt_names=(), engine="h5netcdf", **kwds):
        specs = nc_specs(
            fn,
            "HYDROGRAPH",
            station_idx=station_idx,
            alt_names=alt_names,
            engine=engine,
        )
        attrs = filter_for(cls, specs, **kwds, data_type="HYDROGRAPH")
        return cls(**attrs)


class HRUState(Record):
    hru_id: int = 1
    data: dict[str, Sym] = Field(default_factory=dict)

    def __str__(self):
        return ",".join(map(str, (self.hru_id,) + tuple(self.data.values())))

    @classmethod
    def parse(cls, sol, names=None):
        idx, *values = re.split(r",|\s+", sol.strip())
        if names is None:
            names = [f"S{i}" for i in range(len(values))]
        return cls(hru_id=idx, data=dict(zip(names, values)))


class HRUStateVariableTable(ListCommand):
    """
    Table of HRU state variables.

    If the HRUState include different attributes, the states will be modified to include all attributes.
    """

    root: Sequence[HRUState]
    _Attributes: Sequence[str] = PrivateAttr(None)

    @model_validator(mode="after")
    def set_attributes(self):
        from itertools import chain

        # Get all attribute names from HRUStates
        names = sorted(list(set(chain(*[tuple(s.data.keys()) for s in self.root])))) or None

        self.root = [HRUState(hru_id=s.hru_id, data={n: s.data.get(n, 0.0) for n in names}) for s in self.root]

        # Store names in private class attribute
        self._Attributes = names
        return self

    @classmethod
    def parse(cls, sol: str):
        from collections import defaultdict

        pat = re.compile(
            r"^:HRUStateVariableTable$\s*:Attributes,(?P<atts>.+)$\s*:Units,.*$\s*(?:("
            r"?P<vars>.+)$)+\s*:EndHRUStateVariableTable",
            re.MULTILINE,
        )
        matches = pat.findall(sol)

        d = defaultdict(dict)
        for m in matches:
            atts = [a.strip() for a in m[0].strip(",").split(",")]
            vals = [re.split(r",|\s+", line.strip(",")) for line in m[1:]]

            for val in vals:
                idx, *values = val
            d[idx].update(dict(zip(atts, values)))

        out = []
        for idx, data in d.items():
            out.append(HRUState(hru_id=idx, data=data))
        return cls(out)


class BasinIndex(Command, coerce_numbers_to_str=True):
    """Initial conditions for a flow segment."""

    # TODO: Check that qout and cie can be written with separating commas.
    sb_id: int = 1
    name: str = "watershed"
    channel_storage: float = Field(0.0, alias="ChannelStorage")
    rivulet_storage: float = Field(0.0, alias="RivuletStorage")
    qout: Sequence[float] = Field((1.0, 0.0, 0.0), alias="Qout")
    qlat: Optional[Sequence[float]] = Field(None, alias="Qlat")
    qin: Optional[Sequence[float]] = Field(None, alias="Qin")

    # model_config = ConfigDict(coerce_numbers_to_str=True)

    @property
    def _template(self):
        return """
         :BasinIndex {sb_id} {name}
         {_commands}
          """

    @classmethod
    @no_type_check
    def parse(cls, s):
        pat = r"""
        :BasinIndex (.+?)
        (.+)
        """
        m = re.search(dedent(pat).strip(), s, re.DOTALL)
        index_name = re.split(r",|\s+", m.group(1).strip())
        rec_values = {"sb_id": index_name[0], "name": index_name[1]}
        for line in m.group(2).strip().splitlines():
            all_values = filter(None, re.split(r",|\s+", line.strip()))
            cmd, *values = all_values
            if cmd in [":ChannelStorage", ":RivuletStorage"]:
                # FIXME: assertions should not be found outside of testing code. Replace with conditional logic.
                assert len(values) == 1  # noqa: S101
                rec_values[cmd[1:]] = float(values[0])
            else:
                rec_values[cmd[1:]] = tuple(values)
        return cls(**rec_values)


class BasinStateVariables(ListCommand):
    root: Sequence[BasinIndex]

    @classmethod
    @no_type_check
    def parse(cls, sol):
        pat = r"""
        :BasinStateVariables
        (.+)
        :EndBasinStateVariables
        """
        m = re.search(dedent(pat).strip(), sol, re.DOTALL)
        bis = filter(None, m.group(1).strip().split(":BasinIndex"))
        bs = [BasinIndex.parse(f":BasinIndex {bi}") for bi in bis]
        return cls(bs)


class SoilClasses(ListCommand):
    """
    SoilClasses command.

    Example::

        :SoilClasses
          :Attributes, %SAND, %CLAY, %SILT, %ORGANIC
          :Units, none, none, none, none
          SAND, 1, 0, 0, 0
          LOAM, 0.5, 0.1, 0.4, 0.4
        :EndSoilClasses
    """

    class SoilClass(Record):
        """SoilClass."""

        name: str
        mineral: Optional[tuple[float, float, float]] = None
        organic: Optional[float] = None

        @field_validator("mineral")
        @classmethod
        def validate_mineral(cls, v):
            """Assert sum of mineral fraction is 1."""
            if v is not None:
                # FIXME: assertions should not be found outside of testing code. Replace with conditional logic.
                assert sum(v) == 1, "Mineral fraction should sum to 1."  # noqa: S101
            return v

        @field_validator("mineral")
        @classmethod
        def validate_mineral_pct(cls, v: tuple):
            if v is not None:
                for x in v:
                    # FIXME: assertions should not be found outside of testing code. Replace with conditional logic.
                    assert (x >= 0) and (  # noqa: S101
                        x <= 1
                    ), "Value should be in [0,1]."
            return v

        @field_validator("organic")
        @classmethod
        def validate_organic_pct(cls, v: float):
            if v is not None:
                # FIXME: assertions should not be found outside of testing code. Replace with conditional logic.
                assert (v >= 0) and (v <= 1), "Value should be in [0,1]."  # noqa: S101
            return v

        def __str__(self):
            if self.mineral is None and self.organic is None:
                return self.name
            else:
                fmt = "{name:<16},{mineral[0]:>18},{mineral[1]:>18},{mineral[2]:>18},{organic:>18}"
                return fmt.format(**self.model_dump())

    root: Sequence[SoilClass] = ()
    _Attributes: Sequence[str] = PrivateAttr(["%SAND", "%CLAY", "%SILT", "%ORGANIC"])
    _Units: Sequence[str] = PrivateAttr(["none", "none", "none", "none"])


class VegetationClass(Record):
    name: str = ""
    max_ht: Sym = 0.0
    max_lai: Sym = 0.0
    max_leaf_cond: Sym = 0.0

    def __str__(self):
        template = "{name:<16},{max_ht:>18},{max_lai:>18},{max_leaf_cond:>18}"
        return template.format(**self.model_dump())


class VegetationClasses(ListCommand):
    root: Sequence[VegetationClass]
    _Attributes: Sequence[str] = PrivateAttr(["MAX_HT", "MAX_LAI", "MAX_LEAF_COND"])
    _Units: Sequence[str] = PrivateAttr(["m", "none", "mm_per_s"])


class LandUseClass(Record):
    name: str = ""
    impermeable_frac: Sym = 0.0
    forest_coverage: Sym = 0.0

    def __str__(self):
        template = "{name:<16},{impermeable_frac:>18},{forest_coverage:>18}"
        return template.format(**self.model_dump())


class LandUseClasses(ListCommand):
    root: Sequence[LandUseClass]
    _Attributes: Sequence[str] = PrivateAttr(["IMPERMEABLE_FRAC", "FOREST_COVERAGE"])
    _Units: Sequence[str] = PrivateAttr(["fract", "fract"])


class TerrainClass(Record):
    name: str
    hillslope_length: Sym
    drainage_density: Sym
    topmodel_lambda: Optional[Sym] = None

    def __str__(self):
        out = f"{self.name:<16},{self.hillslope_length:>18},{self.drainage_density:>18}"
        if self.topmodel_lambda is not None:
            out += f",{self.topmodel_lambda:>18}"
        return out


class TerrainClasses(ListCommand):
    root: Sequence[TerrainClass]
    _Attributes: Sequence[str] = PrivateAttr(["HILLSLOPE_LENGTH", "DRAINAGE_DENSITY", "TOPMODEL_LAMBDA"])
    _Units: Sequence[str] = PrivateAttr(["m", "km/km2"])


class _MonthlyRecord(Record):
    name: str = "[DEFAULT]"
    values: Sequence[float] = 12 * [1.0]

    def __str__(self):
        return f"{self.name:<16} " + ", ".join(map(str, self.values))


class SeasonalRelativeLAI(ListCommand):
    root: Sequence[_MonthlyRecord] = (_MonthlyRecord(),)


class SeasonalRelativeHeight(ListCommand):
    root: Sequence[_MonthlyRecord] = (_MonthlyRecord(),)


class SubBasinProperty(Record):
    sb_id: str
    values: Sequence[Sym]

    def __str__(self):
        return indent(
            ", ".join(
                [
                    self.sb_id,
                ]
                + list(map(str, self.values))
            ),
            "  ",
        )

    model_config = ConfigDict(coerce_numbers_to_str=True)


class SubBasinProperties(Command):
    parameters: Optional[Sequence[options.SubBasinProperties]] = Field(None, alias="Parameters")
    records: Optional[Sequence[SubBasinProperty]] = Field(None)


class SoilParameterList(GenericParameterList):
    parameters: Optional[Sequence[options.SoilParameters]] = Field(None, alias="Parameters")


class VegetationParameterList(GenericParameterList):
    parameters: Optional[Sequence[options.VegetationParameters]] = Field(None, alias="Parameters")


class LandUseParameterList(GenericParameterList):
    parameters: Optional[Sequence[options.LandUseParameters]] = Field(None, alias="Parameters")


class EnsembleMode(LineCommand):
    mode: Literal["ENSEMBLE_ENKF"] = "ENSEMBLE_ENKF"
    n: int = Field(..., description="Number of members")


class ObservationErrorModel(LineCommand):
    state: Literal["STREAMFLOW"]
    dist: Literal["DIST_UNIFORM", "DIST_NORMAL", "DIST_GAMMA"]
    p1: float
    p2: float
    adj: Literal["ADDITIVE", "MULTIPLICATIVE"]


class ForcingPerturbation(LineCommand):
    forcing: options.Forcings
    dist: Literal["DIST_UNIFORM", "DIST_NORMAL", "DIST_GAMMA"]
    p1: float
    p2: float
    adj: Literal["ADDITIVE", "MULTIPLICATIVE"]
    hru_grp: str = ""


class AssimilatedState(LineCommand):
    state: Union[options.StateVariables, Literal["STREAMFLOW"]]
    group: str


class AssimilateStreamflow(LineCommand, coerce_numbers_to_str=True):
    """Subbasin ID to assimilate streamflow for."""

    sb_id: str


# TODO: Harmonize this...
# Aliases for convenience
SC = SoilClasses.SoilClass
LU = LandUseClass
VC = VegetationClass
TC = TerrainClass
SB = SubBasin
SP = SoilProfile
PL = ParameterList



from enum import Enum
from typing import Literal


Forcings = Literal[
    "AIR_DENS",
    "AIR_PRES",
    "CLOUD_COVER",
    "DAY_ANGLE",
    "DAY_LENGTH",
    "ET_RADIA",
    "LW_INCOMING",
    "LW_RADIA_NET",
    "OW_PET",
    "PET",
    "PET_MONTH_AVE",
    "POTENTIAL_MELT",
    "PRECIP",
    "PRECIP_5DAY",
    "PRECIP_DAILY_AVE",
    "RAINFALL",
    "RECHARGE",
    "REL_HUMIDITY",
    "SHORTWAVE",
    "SNOWFALL",
    "SNOW_FRAC",
    "SUBDAILY_CORR",
    "SW_RADIA",
    "SW_RADIA_NET",
    "TEMP_AVE",
    "TEMP_AVE_UNC",
    "TEMP_DAILY_AVE",
    "TEMP_DAILY_MAX",
    "TEMP_DAILY_MIN",
    "TEMP_MAX",
    "TEMP_MAX_UNC",
    "TEMP_MIN",
    "TEMP_MIN_UNC",
    "TEMP_MONTH_AVE",
    "TEMP_MONTH_MAX",
    "TEMP_MONTH_MIN",
    "WIND_VEL",
]

# Allowed soil parameters
SoilParameters = Literal[
    "AIR_ENTRY_PRESSURE",
    "ALBEDO_DRY",
    "ALBEDO_WET",
    "BASEFLOW_COEFF",
    "BASEFLOW_COEFF2",
    "BASEFLOW_N",
    "BASEFLOW_THRESH",
    "BF_LOSS_FRACTION",
    "BULK_DENSITY",
    "B_EXP",
    "CLAPP N,CLAPP M",
    "CLAPP_B",
    "CLAY_CON",
    "EVAP_RES_FC",
    "FIELD_CAPACITY",
    "GR4J_X2",
    "GR4J_X3",
    "HBV_BETA",
    "HEAT_CAPACITY",
    "HYDRAUL_COND",
    "INTERFLOW_COEF",
    "MAX_BASEFLOW_RATE",
    "MAX_CAP_RISE_RATE",
    "MAX_INTERFLOW_RATE",
    "MAX_PERC_RATE",
    "ORG_CON",
    "PERC_COEFF",
    "PERC_N",
    "PET_CORRECTION",  # Not in Raven Docs Table A.3
    "POROSITY",
    "SAC_PERC_ALPHA",
    "SAC_PERC_EXPON",
    "SAC_PERC_PFREE",
    "SAND_CON",
    "SAT_RES",
    "SAT_WILT",
    "SHUTTLEWORTH_B",
    "SILT_CON",
    "STONE_FRAC",
    "STORAGE_THRESHOLD",
    "THERMAL_COND",
    "UBC_EVAL_SOIL_DEF",
    "UBC_INFIL_SOIL_DEF",
    "UNAVAIL_FRAC",
    "VIC ALPHA",
    "VIC_EVAP_GAMMA",
    "VIC_ZMAX",
    "VIZ_ZMIN",
    "WETTING_FRONT_PSI",
    "WILTING_PRESSURE",
]

StateVariables = Literal[
    "ATMOSPHERE",
    "ATMOS_PRECIP",
    "CANOPY",
    "CANOPY_SNOW",
    "CANOPY_TEMP",
    "COLD_CONTENT",
    "CONSTITUENT",
    "CONSTITUENT_SINK",
    "CONSTITUENT_SRC",
    "CONSTITUENT_SW",
    "CONVOLUTION",
    "CONV_STOR",
    "CROP_HEAT_UNITS",
    "CUM_INFIL",
    "CUM_SNOWMELT",
    "DEPRESSION",
    "GLACIER",
    "GLACIER_CC",
    "GLACIER_ICE",
    "GROUNDWATER",
    "LAKE_STORAGE",
    "MULTIPLE",
    "PERMAFROST_DEPTH",
    "PONDED_WATER",
    "ROOT",
    "SNOW",
    "SNOW_AGE",
    "SNOW_ALBEDO",
    "SNOW_COVER",
    "SNOW_DEPTH",
    "SNOW_LIQ",
    "SNOW_TEMP",
    "SOIL",
    "SOIL[0]",
    "SOIL[1]",
    "SOIL[2]",
    "SOIL_TEMP",
    "SURFACE_WATER",
    "SURFACE_WATER_TEMP",
    "TOTAL_SWE",
    "TRUNK",
    "WETLAND",
]

LandUseParameters = Literal[
    "ABST_PERCENT",
    "AET_COEFF",
    "BF_LOSS_FRACTION",
    "B_EXP",
    "CC_DECAY_COEFF",
    "DD_AGGRADATION",
    "DD_MELT_TEMP",
    "DD_REFREEZE_TEMP",
    "DEP_K",
    "DEP_MAX",
    "DEP_MAX_FLOW",
    "DEP_N",
    "DEP_SEEP_K",
    "DEP_THRESHOLD",
    "FOREST_COVERAGE",
    "FOREST_PET_CORR",
    "FOREST_SPARSENESS",
    "GAMMA_SCALE",
    "GAMMA_SCALE2",
    "GAMMA_SHAPE",
    "GAMMA_SHAPE2",
    "GLAC_STORAGE_COEFF",
    "GR4J_X4",
    "HBV_GLACIER_AG",
    "HBV_GLACIER_KMIN",
    "HBV_MELT_ASP_CORR",
    "HBV_MELT_FOR_CORR",
    "HBV_MELT_GLACIER_CORR",
    "HMETS_RUNOFF_COEFF",
    "IMPERMEABLE_FRAC",
    "LAKE_PET_CORR",
    "LAKE_REL_COEFF",
    "MAX_DEP_AREA_FRAC",
    "MAX_MELT_FACTOR",
    "MAX_SAT_AREA_FRAC",
    "MELT_FACTOR",
    "MIN_MELT_FACTOR",
    "OW_PET_CORR",
    "PARTITION_COEFF",
    "PDMROF_B",
    "PONDED_EXP",
    "REFREEZE_EXP",
    "REFREEZE_FACTOR",
    "ROUGHNESS",
    "SCS_CN",
    "SCS_IA_FRACTION",
    "SNOW_PATCH_LIMIT",
    "STREAM_FRACTION",
    "UBC_ICEPT_FACTOR",
]

VegetationParameters = Literal[
    "ALBEDO",
    "ALBEDO_WET",
    "CAP_LAI_RATIO",
    "CHU_MATURITY",
    "DRIP_PROPORTION",
    "MAX_CAPACITY",
    "MAX_HEIGHT",
    "MAX_INTERCEPT_RATE",
    "MAX_LAI",
    "MAX_LEAF_COND",
    "MAX_ROOT_LENGTH",
    "MAX_SNOW_CAPACITY",
    "MIN_RESISTIVITY",
    "PET_VEG_CORR",
    "PSI_CRITICAL",
    "RAIN_ICEPT_FACT",
    "RAIN_ICEPT_PCT",
    "RELATIVE_HT",
    "RELATIVE_LAI",
    "ROOTRADIUS",
    "ROOT_EXTINCT",
    "SAI_HT_RATIO",
    "SNOCAP_LAI_RATIO",
    "SNOW_ICEPT_FACT",
    "SNOW_ICEPT_PCT",
    "STEMFLOW_FRAC",
    "SVF_EXTINCTION",
    "TFRAIN",
    "TFSNOW",
    "TRUNK_FRACTION",
    "VEG_DENS",
    "VEG_DIAM",
    "VEG_MBETA",
    "XYLEM_FRAC",
]


class AirPressureMethod(Enum):
    BASIC = "AIRPRESS_BASIC"  # Default
    CONST = "AIRPRESS_CONST"
    DATA = "AIRPRESS_DATA"
    UBC = "AIRPRESS_UBC"


class Calendar(Enum):
    ALL_LEAP = "ALL_LEAP"
    GREGORIAN = "GREGORIAN"
    JULIAN = "JULIAN"
    NOLEAP = "NOLEAP"
    PROLEPTIC_GREGORIAN = "PROLEPTIC_GREGORIAN"
    STANDARD = "STANDARD"
    _365_DAY = "365_DAY"
    _366_DAY = "366_DAY"


class CatchmentRoute(Enum):
    """:CatchmentRoute"""

    DELAYED_FIRST_ORDER = "ROUTE_DELAYED_FIRST_ORDER"
    DUMP = "ROUTE_DUMP"
    EXP = "ROUTE_EXPONENTIAL"
    EXPONENTIAL_UH = "EXPONENTIAL_UH"
    GAMMA = "ROUTE_GAMMA_CONVOLUTION"
    RESERVOIR = "ROUTE_RESERVOIR_SERIES"
    TRI = "ROUTE_TRI_CONVOLUTION"
    TRIANGULAR_UH = "TRIANGULAR_UH"
    TRI_CONVOLUTION = "TRI_CONVOLUTION"


class CloudCoverMethod(Enum):
    NONE = "CLOUDCOV_NONE"  # default
    DATA = "CLOUDCOV_DATA"  # gauge or gridded time series used
    UBC = "CLOUDCOV_UBC"


class EvaluationMetrics(Enum):
    ABSERR = "ABSERR"
    ABSERR_RUN = "ABSERR_RUN"
    ABSMAX = "ABSMAX"
    DIAG_SPEARMAN = "DIAG_SPEARMAN"
    FUZZY_NASH = "FUZZY_NASH"
    KGE_PRIME = "KGE_PRIME"
    KLING_GUPTA = "KLING_GUPTA"
    LOG_NASH = "LOG_NASH"
    NASH_SUTCLIFFE = "NASH_SUTCLIFFE"
    NSC = "NSC"
    PCT_BIAS = "PCT_BIAS"
    PDIFF = "PDIFF"
    RCOEFF = "RCOEFF"
    RMSE = "RMSE"
    TMVOL = "TMVOL"


evaluation_metrics_multiplier = dict(
    ABSERR=-1,
    ABSMAX=-1,
    DIAG_SPEARMAN=1,
    KLING_GUPTA=1,
    LOG_NASH=1,
    NASH_SUTCLIFFE=1,
    NSC=1,
    PCT_BIAS="Not Supported",
    PDIFF=-1,
    RCOEFF=-1,
    RMSE=-1,
    TMVOL=-1,
)


class Evaporation(Enum):
    CONSTANT = "PET_CONSTANT"
    DATA = "PET_DATA"
    FROMMONTHLY = "PET_FROMMONTHLY"
    HAMON_1961 = "PET_HAMON_1961"
    HARGREAVES = "PET_HARGREAVES"
    HARGREAVES_1985 = "PET_HARGREAVES_1985"
    MAKKINK_1957 = "PET_MAKKINK_1957"
    MOHYSE = "PET_MOHYSE"
    MONTHLY_FACTOR = "PET_MONTHLY_FACTOR"
    OUDIN = "PET_OUDIN"
    PENMAN_COMBINATION = "PET_PENMAN_COMBINATION"
    PENMAN_MONTEITH = "PET_PENMAN_MONTEITH"
    PRIESTLEY_TAYLOR = "PET_PRIESTLEY_TAYLOR"
    TURC_1961 = "PET_TURC_1961"
    VAP_DEFICIT = "PET_VAPDEFICIT"


class LWIncomingMethod(Enum):
    DATA = "LW_INC_DATA"
    DEFAULT = "LW_INC_DEFAULT"
    DINGMAN = "LW_INC_DINGMAN"
    SICART = "LW_INC_SICART"
    SKYVIEW = "LW_INC_SKYVIEW"


class Interpolation(Enum):
    AVERAGE_ALL = "INTERP_AVERAGE_ALL"
    FROM_FILE = "INTERP_FROM_FILE"
    INVERSE_DISTANCE = "INTERP_INVERSE_DISTANCE"
    NEAREST_NEIGHBOR = "INTERP_NEAREST_NEIGHBOR"


class LWRadiationMethod(Enum):
    DATA = "LW_RAD_DATA"
    DEFAULT = "LW_RAD_DEFAULT"
    UBCWM = "LW_RAD_UBC"


class MonthlyInterpolationMethod(Enum):
    LINEAR_21 = "MONTHINT_LINEAR_21"
    LINEAR_FOM = "MONTHINT_LINEAR_FOM"
    LINEAR_MID = "MONTHINT_LINEAR_MID"
    UNIFORM = "MONTHINT_UNIFORM"


class OroPETCorrect(Enum):
    HBV = "OROCORR_HBV"
    NONE = "OROCORR_NONE"
    SIMPLELAPSE = "OROCORR_SIMPLELAPSE"


class OroPrecipCorrect(Enum):
    HBV = "OROCORR_HBV"
    NONE = "OROCORR_NONE"
    SIMPLELAPSE = "OROCORR_SIMPLELAPSE"
    UBC = "OROCORR_UBC"


class OroTempCorrect(Enum):
    HBV = "OROCORR_HBV"
    NONE = "OROCORR_NONE"
    SIMPLELAPSE = "OROCORR_SIMPLELAPSE"
    UBC = "OROCORR_UBC"
    UBC2 = "OROCORR_UBC_2"


class PotentialMeltMethod(Enum):
    """:PotentialMelt algorithms"""

    DATA = "POTMELT_DATA"
    DEGREE_DAY = "POTMELT_DEGREE_DAY"
    EB = "POTMELT_EB"
    HBV = "POTMELT_HBV"
    HMETS = "POTMELT_HMETS"
    NONE = "POTMELT_NONE"
    RESTRICTED = "POTMELT_RESTRICTED"
    UBC = "POTMELT_UBC"
    USACE = "POTMELT_USACE"


class Precipitation(Enum):
    DEFAULT = "PRECIP_RAVEN"


class PrecipIceptFract(Enum):
    USER = "PRECIP_ICEPT_USER"  # default
    EXPLAI = "PRECIP_ICEPT_EXPLAI"
    HEDSTROM = "PRECIP_ICEPT_HEDSTROM"
    LAI = "PRECIP_ICEPT_LAI"
    NONE = "PRECIP_ICEPT_NONE"


class RainSnowFraction(Enum):
    DATA = "RAINSNOW_DATA"
    DINGMAN = "RAINSNOW_DINGMAN"
    HARDER = "RAINSNOW_HARDER"
    HBV = "RAINSNOW_HBV"
    HSPF = "RAINSNOW_HSPF"
    SNTHERM89 = "RAINSNOW_SNTHERM89"
    UBC = "RAINSNOW_UBC"
    WANG = "RAINSNOW_WANG"


class RechargeMethod(Enum):
    DATA = "RECHARGE_DATA"
    NONE = "RECHARGE_NONE"


class RelativeHumidityMethod(Enum):
    CONSTANT = "RELHUM_CONSTANT"
    CORR = "RELHUM_CORR"
    DATA = "RELHUM_DATA"
    MINDEWPT = "RELHUM_MINDEWPT"
    WINDVEL = "WINDVEL_CORR"


class Routing(Enum):
    DIFFUSIVE_WAVE = "ROUTE_DIFFUSIVE_WAVE"
    HYDROLOGIC = "ROUTE_HYDROLOGIC"
    MUSKINGUM = "MUSKINGUM"
    NONE = "ROUTE_NONE"
    PLUG_FLOW = "ROUTE_PLUG_FLOW"
    STORAGE_COEFF = "ROUTE_STORAGE_COEFF"


class SoilModel(Enum):
    ONE_LAYER = "SOIL_ONE_LAYER"
    TWO_LAYER = "SOIL_TWO_LAYER"
    MULTILAYER = "SOIL_MULTILAYER"


SubBasinProperties = Literal[
    "CELERITY",
    "DIFFUSIVITY",
    "GAMMA_SCALE",
    "GAMMA_SHAPE",
    "MANNINGS_N",
    "NUM_RESERVOIRS",
    "Q_REFERENCE",
    "RAIN_CORR",
    "RES_CONSTANT",
    "SLOPE",
    "SNOW_CORR",
    "TIME_CONC",
    "TIME_LAG",
    "TIME_TO_PEAK",
]


class SubdailyMethod(Enum):
    NONE = "SUBDAILY_NONE"
    SIMPLE = "SUBDAILY_SIMPLE"
    UBC = "SUBDAILY_UBC"


class SWCanopyCorrect(Enum):
    NONE = "SW_CANOPY_CORR_NONE"  # Default
    DYNAMIC = "SW_CANOPY_CORR_DYNAMIC"
    STATIC = "SW_CANOPY_CORR_STATIC"
    UBC = "SW_CANOPY_CORR_UBC"


class SWCloudCorrect(Enum):
    NONE = "SW_CLOUD_CORR_NONE"  # Default
    ANNANDALE = "SW_CLOUD_CORR_ANNANDALE"
    DINGMAN = "SW_CLOUD_CORR_DINGMAN"
    UBC = "SW_CLOUD_CORR_UBCWM"


class SWRadiationMethod(Enum):
    DATA = "SW_RAD_DATA"
    DEFAULT = "SW_RAD_DEFAULT"
    UBCWM = "SW_RAD_UBCWM"


class WindspeedMethod(Enum):
    CONSTANT = "WINDVEL_CONSTANT"
    DATA = "WINDVEL_DATA"
    UBC = "WINDVEL_UBC"


class EnKFMode(Enum):
    CLOSED_LOOP = "ENKF_CLOSED_LOOP"
    FORECAST = "ENKF_FORECAST"
    OPEN_FORECAST = "ENKF_OPEN_FORECAST"
    OPEN_LOOP = "ENKF_OPEN_LOOP"
    SPINUP = "ENKF_SPINUP"



"""Publishing utilities for RavenPy."""

from __future__ import annotations
import os
import re
from io import StringIO
from pathlib import Path
from typing import TextIO


def publish_release_notes(
    style: str = "md",
    file: os.PathLike[str] | StringIO | TextIO | None = None,
    changes: str | os.PathLike[str] | None = None,
) -> str | None:
    """
    Format release notes in Markdown or ReStructuredText.

    Parameters
    ----------
    style : {"rst", "md"}
        Use ReStructuredText formatting or Markdown. Default: Markdown.
    file : {os.PathLike, StringIO, TextIO}, optional
        If provided, prints to the given file-like object. Otherwise, returns a string.
    changes : str or os.PathLike[str], optional
        If provided, manually points to the file where the changelog can be found.
        Assumes a relative path otherwise.

    Returns
    -------
    str, optional
        If `file` not provided, the formatted release notes.

    Notes
    -----
    This function is used solely for development and packaging purposes.
    """
    if isinstance(changes, str | Path):
        changes_file = Path(changes).absolute()
    else:
        changes_file = Path(__file__).absolute().parents[3].joinpath("CHANGELOG.rst")

    if not changes_file.exists():
        raise FileNotFoundError("Changelog file not found in RavenPy folder tree.")

    with Path(changes_file).open(encoding="utf-8") as hf:
        changes = hf.read()

    if style == "rst":
        hyperlink_replacements = {
            r":issue:`([0-9]+)`": r"`GH/\1 <https://github.com/CSHS-CWRA/RavenPy/issues/\1>`_",
            r":pull:`([0-9]+)`": r"`PR/\1 <https://github.com/CSHS-CWRA/RavenPy/pull/\>`_",
            r":user:`([a-zA-Z0-9_.-]+)`": r"`@\1 <https://github.com/\1>`_",
        }
    elif style == "md":
        hyperlink_replacements = {
            r":issue:`([0-9]+)`": r"[GH/\1](https://github.com/CSHS-CWRA/RavenPy/issues/\1)",
            r":pull:`([0-9]+)`": r"[PR/\1](https://github.com/CSHS-CWRA/RavenPy/pull/\1)",
            r":user:`([a-zA-Z0-9_.-]+)`": r"[@\1](https://github.com/\1)",
        }
    else:
        msg = f"Formatting style not supported: {style}"
        raise NotImplementedError(msg)

    for search, replacement in hyperlink_replacements.items():
        changes = re.sub(search, replacement, changes)

    if style == "md":
        changes = changes.replace("=========\nChangelog\n=========", "# Changelog")

        titles = {r"\n(.*?)\n([\-]{1,})": "-", r"\n(.*?)\n([\^]{1,})": "^"}
        for title_expression, level in titles.items():
            found = re.findall(title_expression, changes)
            for grouping in found:
                fixed_grouping = str(grouping[0]).replace("(", r"\(").replace(")", r"\)")
                search = rf"({fixed_grouping})\n([\{level}]{'{' + str(len(grouping[1])) + '}'})"
                replacement = f"{'##' if level == '-' else '###'} {grouping[0]}"
                changes = re.sub(search, replacement, changes)

        link_expressions = r"[\`]{1}([\w\s]+)\s<(.+)>`\_"
        found = re.findall(link_expressions, changes)
        for grouping in found:
            search = rf"`{grouping[0]} <.+>`\_"
            replacement = f"[{str(grouping[0]).strip()}]({grouping[1]})"
            changes = re.sub(search, replacement, changes)

    if not file:
        return changes
    if isinstance(file, Path | os.PathLike):
        with Path(file).open("w", encoding="utf-8") as f:
            print(changes, file=f)
    else:
        print(changes, file=file)
    return None



import tempfile
import warnings
from collections.abc import Sequence
from dataclasses import astuple
from pathlib import Path
from typing import Optional, Union

from spotpy.parameter import Uniform, generate

from ravenpy import Emulator
from ravenpy.config.base import Params
from ravenpy.config.options import evaluation_metrics_multiplier
from ravenpy.config.rvs import Config


class SpotSetup:
    def __init__(
        self,
        config: Config,
        low: Union[Params, Sequence],
        high: Union[Params, Sequence],
        workdir: Optional[Union[Path, str]] = None,
    ):
        """
        Class to configure spotpy with Raven emulators.

        Parameters
        ----------
        config : Config
            Emulator Config instance with symbolic expressions.
        low : Union[Params, Sequence]
            Lower boundary for parameters.
        high : Union[Params, Sequence]
            Upper boundary for parameters.
        workdir : Union[str, Path], optional
            Work directory. If None, a temporary directory will be created.
        """
        if not config.is_symbolic:
            raise ValueError("config should be a symbolic configuration, where params are not set to their numerical values.")

        self.config = config
        self.path = Path(workdir or tempfile.mkdtemp())
        self.diagnostics = None

        if config.suppress_output is not True:
            warnings.warn("Add the `SuppressOutput` command to the configuration to reduce IO.", stacklevel=2)

        if config.evaluation_metrics is None:
            raise AttributeError(":EvaluationMetrics is undefined.")

        # Get evaluation metrics and their multiplier (spotpy maximizes the obj function)
        self.metrics = [m.value for m in config.evaluation_metrics]
        self._multipliers = {m: evaluation_metrics_multiplier[m] for m in self.metrics}

        p = config.params

        self.pnames = list(p.__dataclass_fields__.keys())
        self.pdist = self.init_params(low, high)
        self._iteration = 0

    def init_params(self, low: Union[Params, Sequence], high: [Params, Sequence]):
        # Validate parameters
        low = astuple(self._to_dataclass(low))
        high = astuple(self._to_dataclass(high))

        pdist = []
        for i in range(len(low)):
            pdist.append(  # noqa: PERF401
                Uniform(self.pnames[i], low=low[i], high=high[i])
            )
        return pdist

    def _to_dataclass(self, p):
        """Validate parameters against Params dataclass from model configuration."""
        kls = self.config.params.__class__
        if isinstance(p, kls):
            return p
        return kls(*p)

    def parameters(self):
        """Return a random parameter combination."""
        return generate(self.pdist)

    @staticmethod
    def evaluation():
        """
        Return the observation.

        Since Raven computes the objective function itself, we simply return a placeholder.
        """
        return 1

    def simulation(self, x):
        """Run the model, but return a placeholder value instead of the model output."""
        self._iteration += 1

        # Update parameters
        c = self.config.set_params(list(x))

        # Create emulator instance
        emulator = Emulator(config=c, workdir=self.path / f"c{self._iteration:03}")

        # Run the model
        output = emulator.run()

        self.diagnostics = output.diagnostics

        self._iteration += 1

        return 1

    # FIXME: Verify that this function is correct
    def objectivefunction(self, evaluation, simulation):  # noqa: F841
        """
        Return the objective function.

        Note that we short-circuit the evaluation and simulation entries, since the objective function has already
        been computed by Raven.
        """
        out = [self.diagnostics[f"DIAG_{m}"][0] * self._multipliers[m] for m in self.metrics]

        return out



import logging
import os
import pathlib
import shutil
import sys

import pytest


def filter_raven():
    # Find the real conda-installed tool path
    real_tool = shutil.which("raven")
    assert real_tool is not None, "raven must be installed for this test"

    # Get the directory the conda tool is in
    conda_tool_dir = pathlib.Path(real_tool).parent

    # Create a new PATH that includes everything *except* the conda env's tool directory
    filtered_path = os.pathsep.join(
        [p for p in os.environ["PATH"].split(os.pathsep) if pathlib.Path(p).resolve() != pathlib.Path(conda_tool_dir).resolve()]
    )

    return filtered_path


@pytest.fixture
def hide_module(monkeypatch):
    def _hide(name):
        monkeypatch.setitem(sys.modules, name, None)

    return _hide


class TestMissing:
    def test_missing_raven_binary(self, monkeypatch, tmpdir, hide_module):
        """Test for behaviour when binary is missing from the system path."""
        # Set up a temporary directory to simulate the absence of the raven binary
        filtered_path = filter_raven()
        monkeypatch.setenv("PATH", f"{tmpdir}{os.pathsep}{filtered_path}")

        # Hide the raven_hydro module
        hide_module("raven_hydro")
        hide_module("raven_hydro._version")
        hide_module("raven_hydro.libraven")

        # Force the raven modules to be reloaded
        del sys.modules["ravenpy"]
        del sys.modules["ravenpy._raven"]
        del sys.modules["ravenpy.config.defaults"]

        # Now the tool should be "missing"
        # if running from tox in a conda environment, the raven binary is likely to be present
        if os.environ.get("CONDA_PREFIX") and os.environ.get("TOX"):
            logging.info("Running in a development conda environment with tox. Raven is expected to be present.")
        else:
            assert shutil.which("raven") is None

        # Loading the module should raise a RuntimeError
        with pytest.raises(RuntimeError):
            import ravenpy  # noqa: F401

        # Check that setting the RAVENPY_RAVEN_BINARY_PATH environment variable works
        monkeypatch.setenv("RAVENPY_RAVEN_BINARY_PATH", "some/path/to/raven")
        import ravenpy

        assert ravenpy.RAVEN_EXEC_PATH == "some/path/to/raven"

        # Check that the raven_hydro library is not imported
        from ravenpy.config.defaults import __raven_version__

        assert __raven_version__ == "0.0.0"



"""Distributed model workflow."""

import datetime as dt

from ravenpy import Emulator
from ravenpy.config import commands as rc
from ravenpy.config.emulators import GR4JCN
from ravenpy.extractors.routing_product import (
    BasinMakerExtractor,
    GridWeightExtractor,
    open_shapefile,
    upstream_from_coords,
)


def test_simple_workflow(tmp_path, yangtze):
    shp_path = yangtze.fetch("basinmaker/drainage_region_0175_v2-1/finalcat_info_v2-1.zip")

    # Note that for this to work, the coordinates must be in the small
    # BasinMaker example (drainage_region_0175)
    df = open_shapefile(shp_path)

    # Gauge station for observations at Matapedia
    # SubId: 175000128
    # -67.12542 48.10417
    sub = upstream_from_coords(-67.12542, 48.10417, df)

    # Confirm we got the right watershed
    assert 175000128 in sub["SubId"].to_list()

    # Extract the subbasins and HRUs (one HRU per sub-basin)
    bm = BasinMakerExtractor(
        df=sub,
        hru_aspect_convention="ArcGIS",
    )
    rvh = bm.extract(hru_from_sb=True)

    # Streamflow obs
    qobs_fn = yangtze.fetch("matapedia/Qobs_Matapedia_01BD009.nc")

    qobs = rc.ObservationData.from_nc(
        qobs_fn,
        uid=175000128,
        alt_names=("discharge",),
    )

    # Meteo obs for GriddedForcing - does not work because subbasins do not overlap 100% with the ERA data
    meteo_grid_fn = yangtze.fetch("matapedia/Matapedia_meteo_data_2D.nc")

    # Dict of GW attributes
    gw = GridWeightExtractor(
        meteo_grid_fn,
        shp_path,
        dim_names=("longitude", "latitude"),
        var_names=("longitude", "latitude"),
        gauge_ids=[
            "01BD009",
        ],
    ).extract()

    assert gw["number_hrus"] == len(sub)

    # Write GW command to file
    gw_fn = tmp_path / "gw.txt"
    gw_fn.write_text(rc.GridWeights(**gw).to_rv())

    forcing = {"TEMP_MIN": "tmin", "TEMP_MAX": "tmax", "PRECIP": "pr"}

    [rc.GriddedForcing.from_nc(meteo_grid_fn, dtyp, alt_names=(alias,), grid_weights=gw_fn) for (dtyp, alias) in forcing.items()]
    # Weights for some HRUs do not sum to one.

    # Meteo forcing per station (virtual stations, since this is ERA5 data)
    meteo_station = yangtze.fetch("matapedia/Matapedia_meteo_data_stations.nc")

    [rc.StationForcing.from_nc(meteo_station, dtyp, alt_names=(alias,)) for (dtyp, alias) in forcing.items()]
    # TODO: Complete with weights calculations

    # Virtual Gauges
    gauges = [
        rc.Gauge.from_nc(
            meteo_station,
            data_type=[s for s in forcing.keys()],
            station_idx=i + 1,
            alt_names=forcing,
        )
        for i in range(6)
    ]

    conf = GR4JCN(
        params=[0.529, -3.396, 407.29, 1.072, 16.9, 0.947],
        StartDate=dt.datetime(2000, 1, 1),
        Duration=15,
        GlobalParameter={"AVG_ANNUAL_RUNOFF": 208.480},
        ObservationData=[qobs],
        Gauge=gauges,
        **rvh,
    )

    out = Emulator(conf, workdir=tmp_path).run()

    # Number of gauged sub-basins
    ng = sum([sb.gauged for sb in conf.sub_basins])
    assert len(out.hydrograph.nbasins) == ng



import os
import typing
from collections.abc import Sequence
from pathlib import Path
from typing import Optional, Union

import cf_xarray  # noqa: F401
import numpy as np
import xarray as xr

from .conventions import CF_RAVEN, MonthlyAverages
from .defaults import RAVEN_NO_DATA_VALUE


def nc_specs(
    fn: Union[str, os.PathLike[str]],
    data_type: str,
    station_idx: Optional[int] = None,
    alt_names: Optional[Union[str, Sequence[str]]] = None,
    mon_ave: bool = False,
    engine: str = "h5netcdf",
):
    """
    Extract specifications from netCDF file.

    Parameters
    ----------
    fn : str, Path
        NetCDF file path or DAP link.
    data_type : str
        Raven data type.
    station_idx : int, optional
        Index along station dimension. Starts at 1.
    alt_names : str or list of str, optional
        Alternative variable names for data type if not the CF standard default.
    mon_ave : bool
        If True, compute the monthly average.
    engine : str
        The engine used to open the dataset. Default is 'h5netcdf'.

    Returns
    -------
    dict

    Notes
    -----
    Attributes: var_name_nc, dim_names_nc, units, linear_transform, latitude_var_name_nc, longitude_var_name_nc,
    elevation_var_name_nc
    latitude, longitude, elevation, name
    """
    from ravenpy.utilities.coords import infer_scale_and_offset

    if isinstance(fn, str) and str(fn)[:4] == "http":
        pass
    elif Path(fn).exists():
        fn = os.path.realpath(fn, strict=True)
    else:
        raise ValueError("NetCDF file not found.")

    if alt_names is None:
        alt_names = ()
    elif isinstance(alt_names, str):
        alt_names = (alt_names,)

    attrs = {
        "file_name_nc": fn,
        "data_type": data_type,
        "forcing_type": data_type,
        "station_idx": station_idx,
    }

    with xr.open_dataset(fn, engine=engine) as ds:
        var_names = CF_RAVEN.get(data_type, ()) + tuple(alt_names)
        if len(var_names) == 0:
            raise ValueError(f"Provide alternative variable name with `alt_names` for {data_type}")

        for v in var_names:
            if v in ds.data_vars:
                nc_var = ds[v]
                attrs["var_name_nc"] = v
                attrs["dim_names_nc"] = nc_var.dims
                attrs["_time_dim_name_nc"] = ds.cf["time"].name
                attrs["_dim_size_nc"] = dict(zip(nc_var.dims, nc_var.shape))
                attrs["units"] = nc_var.attrs.get("units")
                if attrs["units"] is not None:
                    s, o = infer_scale_and_offset(nc_var, data_type)
                    attrs["linear_transform"] = dict(scale=s, offset=o)
                if mon_ave:
                    ma = MonthlyAverages.get(data_type)
                    if ma:
                        attrs[ma] = nc_var.groupby("time.month").mean().values.tolist()

                break
        else:
            raise ValueError(f"No variable found for {data_type}.\n {ds.data_vars}")

        if station_idx is not None:
            # Convert to NumPy 0-based indexing
            i = station_idx - 1

            try:
                attrs["latitude_var_name_nc"] = ds.cf["latitude"].name
                attrs["longitude_var_name_nc"] = ds.cf["longitude"].name

                attrs["latitude"] = ds.cf["latitude"][i]
                attrs["longitude"] = ds.cf["longitude"][i]

            except KeyError:  # noqa: S110
                pass

            try:
                nc_elev = ds.cf["vertical"].name
            except KeyError:
                nc_elev = "elevation" if "elevation" in ds else None

            finally:
                if nc_elev is not None:
                    attrs["elevation_var_name_nc"] = nc_elev
                    try:
                        attrs["elevation"] = ds[nc_elev][i]
                    except IndexError:
                        # Elevation is a scalar
                        attrs["elevation"] = ds[nc_elev].item(0)

            if "station_id" in ds:
                if ds["station_id"].shape and len(ds["station_id"]) > i:
                    attrs["name"] = ds["station_id"].values[i]

    return attrs


def filter_for(kls, attrs, **kwds):
    """
    Return attributes that are fields of dataclass.

    Notes
    -----
    If attrs includes an attribute name and its Raven alias, e.g. `linear_transform` and `LinearTransform`, the latter will have priority.
    """
    from pydantic._internal._model_construction import ModelMetaclass

    from .commands import Command

    attrs.update(kwds)

    if hasattr(kls, "__dataclass_fields__"):
        return {k: v for (k, v) in attrs.items() if k in kls.__dataclass_fields__.keys()}
    else:
        out = {}
        for key, field in kls.model_fields.items():
            if field.alias in attrs:
                out[field.alias] = attrs[field.alias]
            elif key in attrs:
                out[key] = attrs[key]
            elif isinstance(field.annotation, ModelMetaclass) and issubclass(field.annotation, Command):
                out[key] = filter_for(field.annotation, attrs)
        return out
        # return {k: v for (k, v) in attrs.items() if k in kls.__fields__.keys()}


def get_average_annual_runoff(
    nc_file_path: Union[str, os.PathLike[str]],
    area_in_m2: float,
    time_dim: str = "time",
    obs_var: str = "qobs",
    na_value: Union[int, float] = RAVEN_NO_DATA_VALUE,
):
    """Compute the average annual runoff from observed data."""
    with xr.open_dataset(nc_file_path) as ds:
        q_obs = ds.where(ds[obs_var] != na_value)[obs_var]
        q_obs *= 86400.0  # convert m**3/s to m**3/d
        axis = q_obs.dims.index(time_dim)
        # avg daily runoff [m3/d] for each year in record
        q_year = np.nanmean(q_obs.groupby("time.year").mean("time"), axis=axis)
        q_year = q_year / area_in_m2 * 365 * 1000.0  # [mm/yr] for each year in record
        q_year = np.mean(q_year)  # [mm/yr] mean over all years in record

    return q_year


def get_annotations(a):
    """Return all annotations inside [] or Union[...]."""
    for arg in typing.get_args(a):
        if typing.get_origin(arg) == Union:
            yield from get_annotations(arg)
        else:
            yield arg



import datetime as dt

from ravenpy import Emulator, EnsembleReader
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.emulators import GR4JCN


# Alternative names for variables in meteo forcing file
alt_names = {
    "RAINFALL": "rain",
    "TEMP_MIN": "tmin",
    "TEMP_MAX": "tmax",
    "SNOWFALL": "snow",
}


# @pytest.mark.xfail()
def test_enkf(salmon_hru, tmp_path, yangtze):
    """Test one run of Ensemble Kalman Filter data assimilation."""
    cls = GR4JCN
    # name = "GR4JCN"
    data_type = ["RAINFALL", "TEMP_MIN", "TEMP_MAX", "SNOWFALL"]

    salmon_file = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")

    conf = cls(
        params=(0.14, -0.005, 576, 7.0, 1.1, 0.92),
        Gauge=[
            rc.Gauge.from_nc(
                salmon_file,
                data_type=data_type,
                alt_names=alt_names,
                data_kwds={"ALL": {"elevation": salmon_hru["land"]["elevation"]}},
            ),
        ],
        ObservationData=[rc.ObservationData.from_nc(salmon_file, alt_names="qobs")],
        HRUs=[salmon_hru["land"]],
        StartDate=dt.datetime(1996, 9, 1),
        EndDate=dt.datetime(1996, 9, 30),
        EnsembleMode=rc.EnsembleMode(n=7),
        EnKFMode=o.EnKFMode.SPINUP,
        RunName="spinup",
        EvaluationMetrics=("NASH_SUTCLIFFE",),
        OutputDirectoryFormat="./ens_*",
        ForcingPerturbation=[
            rc.ForcingPerturbation(
                forcing="PRECIP",
                dist="DIST_NORMAL",
                p1=1,
                p2=0.07,
                adj="MULTIPLICATIVE",
            ),
        ],
        DefineHRUGroups=["All"],
        HRUGroup=[{"name": "All", "groups": ["1"]}],
        AssimilatedState=[
            rc.AssimilatedState(state="SOIL[0]", group="All"),
            rc.AssimilatedState(state="SOIL[1]", group="All"),
        ],
        AssimilateStreamflow=[rc.AssimilateStreamflow(sb_id=1)],
        ObservationErrorModel=[
            rc.ObservationErrorModel(
                state="STREAMFLOW",
                dist="DIST_NORMAL",
                p1=1,
                p2=0.07,
                adj="MULTIPLICATIVE",
            )
        ],
    )

    # Spinup
    Emulator(config=conf, workdir=tmp_path, overwrite=True).run(overwrite=True)

    # Closed Loop
    conf_loop = conf.duplicate(
        EnKFMode=o.EnKFMode.CLOSED_LOOP,
        RunName="loop",
        SolutionRunName="spinup",
        UniformInitialConditions=None,
        StartDate=dt.datetime(1996, 9, 30),
        EndDate=dt.datetime(1996, 10, 15),
    )

    # Loop
    Emulator(config=conf_loop, workdir=tmp_path).run()

    # Forecast
    conf_cast = conf_loop.duplicate(EnKFMode=o.EnKFMode.FORECAST, RunName="forecast", SolutionRunName="loop")

    # Cast
    Emulator(config=conf_cast, workdir=tmp_path).run()

    paths = list(tmp_path.glob("ens_*"))
    paths.sort()
    assert len(paths) == 7

    ens = EnsembleReader(run_name=conf_cast.run_name, paths=paths)
    assert conf_cast.run_name == "forecast"
    assert len(ens.files["hydrograph"]) == 7



from ravenpy.config.emulators.blended import Blended
from ravenpy.config.emulators.canadianshield import CanadianShield
from ravenpy.config.emulators.gr4jcn import GR4JCN
from ravenpy.config.emulators.hbvec import HBVEC
from ravenpy.config.emulators.hmets import HMETS
from ravenpy.config.emulators.hypr import HYPR
from ravenpy.config.emulators.mohyse import Mohyse
from ravenpy.config.emulators.routing import BasicRoute
from ravenpy.config.emulators.sacsma import SACSMA


__all__ = [
    "get_model",
]


def get_model(name):
    """
    Return the corresponding Raven emulator configuration class.

    Parameters
    ----------
    name : str
        Model class name or model identifier.

    Returns
    -------
    Raven model configuration class
    """
    cls = globals().get(name)

    if cls is None:
        raise ValueError(f"Model {name} is not recognized.")

    return cls



import collections
import re
from pathlib import Path
from typing import Optional, Union

import cftime
from xarray import open_dataset

from ..config import commands as rc
from .conventions import RAVEN_OUTPUT_FMT


def parse_rv(rv: str, command: str) -> str:
    """
    Parse an RV file to find the value of a Command (one liner).

    Parameters
    ----------
    rv : str
        The Raven command string, e.g. "RunName".
    command : str
        Valid Raven command.

    Returns
    -------
    str:
        Command value. None if not found.
    """
    pat = re.compile(rf":{command}\s+(.+)")
    if match := re.search(pat, rv):
        return match.groups()[0]


def parse_diagnostics(fn: Path):
    """Return dictionary of performance metrics."""
    import csv

    if fn.exists():
        out = collections.defaultdict(list)

        with Path(fn).open() as f:
            reader = csv.reader(f.readlines())
            header = next(reader)
            for row in reader:
                for key, val in zip(header, row):
                    if "DIAG" in key:
                        val = float(val)
                    out[key].append(val)

            out.pop("")

        return out

    else:
        raise FileNotFoundError(fn)


def parse_nc(fn: Path):
    """Open netCDF dataset with xarray if the path is valid, otherwise return None."""
    if fn.exists():
        if fn.suffix == ".nc":
            return open_dataset(fn)
        else:
            raise NotImplementedError
    else:
        raise FileNotFoundError(fn)


def parse_solution(fn: Path, calendar: str = "PROLEPTIC_GREGORIAN"):
    """Return command objects from the model output `solution.rvc`."""
    if fn.exists():
        if fn.suffix == ".rvc":
            solution = fn.read_text()

            out = {
                "hru_state_variable_table": rc.HRUStateVariableTable.parse(solution),
                "basin_state_variables": rc.BasinStateVariables.parse(solution),
                "start_date": _time_stamp_from_solution(solution, calendar),
            }
            return out

        else:
            raise NotImplementedError
    else:
        raise FileNotFoundError(fn)


def parse_raven_messages(path):
    """Parse Raven_errors and extract the messages, structured by types."""
    messages = {
        "ERROR": [],
        "WARNING": [],
        "ADVISORY": [],
        "SIMULATION COMPLETE": False,
    }

    # The error message for an unknown command is exceptionally on two lines
    # (the second starts with a triple space)
    for m in re.findall("^([A-Z ]+) :(.+)(?:\n {3}(.+))?", path.read_text(), re.M):
        if m[0] == "SIMULATION COMPLETE":
            messages["SIMULATION COMPLETE"] = True
            continue
        msg_type = str(m[0]).strip()
        msg = f"{m[1]} {m[2]}".strip()
        if msg == "Errors found in input data. See Raven_errors.txt for details":
            # Skip this one because it's a bit circular
            continue
        messages[msg_type].append(msg)

    return messages


def output_files(run_name: str, path: Path):
    """Return path to each output file if it exists."""
    out = {}
    for k, v in RAVEN_OUTPUT_FMT.items():
        p = path / v.format(run_name=run_name + "_" if run_name is not None else "")
        if p.exists():
            out[k] = p
    return out


def parse_outputs(run_name: str, outputdir: Optional[Union[str, Path]] = None):
    """
    Parse outputs from model execution.

    Parameters
    ----------
    run_name : str
        RunName value identifying model outputs.
    outputdir : str or Path
        Path to model output directory. Current directory if None.

    Returns
    -------
    dict
      Dictionary holding model outputs:
        - hydrograph: xarray.Dataset
        - storage: xarray.Dataset
        - solution: Dict[str, Command]
        - diagnostics: Dict[str, list]

    Notes
    -----
    Values are set to None if no file is found.
    """
    if outputdir is None:
        outputdir = Path.cwd()
    elif isinstance(type(outputdir), str):
        outputdir = Path(outputdir)

    parser = RAVEN_OUTPUT_PARSERS

    out = {}
    files = output_files(run_name, outputdir)

    for key, path in files.items():
        if path.exists():
            if key in parser:
                out[key] = parser[key](path)
            else:
                out[key] = path.read_text()

    return out


def _time_stamp_from_solution(solution: str, calendar: str) -> Optional[cftime.datetime]:
    """Return datetime from solution TimeStamp."""
    match = re.search(r":TimeStamp (\S+ \S+)", solution)
    if match:
        tt = cftime._parse_date(match.groups()[0])
        return cftime.datetime(*tt, calendar=calendar)


RAVEN_OUTPUT_PARSERS = {
    "solution": parse_solution,
    "diagnostics": parse_diagnostics,
    "storage": parse_nc,
    "hydrograph": parse_nc,
    "messages": parse_raven_messages,
}



import pathlib
from importlib.util import find_spec

from ravenpy import Emulator, EnsembleReader


def test_ensemble_reader(gr4jcn_config, tmp_path):
    """Mimics code in docs/outputs.md"""
    conf, params = gr4jcn_config

    params = [params, params]

    # Output directory for all simulations
    p = tmp_path / "ensemble"

    # Writing to the same custom output cause segfaults
    # conf.custom_output = [rc.CustomOutput(time_per="DAILY", stat="AVERAGE", variable="RAINFALL",
    # space_agg="ENTIRE_WATERSHED")]

    # Run the model for each parameter set in `params`
    runs = [Emulator(conf.set_params(param), workdir=p / f"m{i}").run() for i, param in enumerate(params)]

    ens = EnsembleReader(runs=runs, dim="parameters")
    assert len(ens.hydrograph.parameters) == 2

    # Create a list of output paths using glob
    paths = p.glob("**/output")

    ens = EnsembleReader(paths=paths, dim="parameters")
    assert len(ens.hydrograph.parameters) == 2


def test_package_metadata():
    """Test the package metadata."""
    project = find_spec("ravenpy")

    assert project is not None
    assert project.submodule_search_locations is not None

    location = project.submodule_search_locations[0]

    metadata = pathlib.Path(location).resolve().joinpath("__init__.py")

    with metadata.open() as f:
        contents = f.read()
        assert """David Huard""" in contents
        assert '__email__ = "huard.david@ouranos.ca"' in contents
        assert '__version__ = "0.19.1"' in contents
